@@ -6,7 +6,7 @@ use crate::{
     header::{
         auth_results::{self, AuthenticationResultsHeader},
         received_spf::ReceivedSpfHeader,
-        ToHeaderField,
+        HeaderField,
     },
     resolver::MockResolver,
     verify::{self, VerificationResult},
@@ -67,7 +67,7 @@ impl ConnectionData {
 /// Message-scoped session data.
 #[derive(Clone, Debug, Eq, Hash, PartialEq)]
 struct MessageData {
-    // At the end of the authorisation process `results` may contain the HELO or
+    // At the end of the authorisation session `results` may contain the HELO or
     // MAIL FROM result, both, or neither (when senders were skipped).
     results: Vec<VerificationResult>,
     auth_results_i: usize,
@@ -121,7 +121,7 @@ impl AuthSession {
     /// 3207, section 4.2.
     pub fn authorize_helo(
         &mut self,
-        context: &(impl ActionContext + SetErrorReply),
+        context: &impl SetErrorReply,
         helo_host: impl Into<String>,
         mock_resolver: Option<&MockResolver>,
     ) -> milter::Result<Status> {
@@ -170,7 +170,7 @@ impl AuthSession {
     /// is established.
     pub fn authorize_mail_from(
         &mut self,
-        context: &(impl ActionContext + SetErrorReply),
+        context: &impl SetErrorReply,
         mail_from: &str,
         mock_resolver: Option<&MockResolver>,
     ) -> milter::Result<Status> {
@@ -258,6 +258,11 @@ impl AuthSession {
                     );
                     message.auth_results_deletions.push(message.auth_results_i);
                 }
+            } else {
+                debug!(
+                    "{}: failed to parse incoming Authentication-Results header at index {}",
+                    id, message.auth_results_i
+                );
             }
         }
     }
@@ -375,12 +380,12 @@ fn to_reply_params<'a>(
     }
 }
 
-// The configuration may supply arbitrary reply text strings. According to
-// libmilter API documentation, `%` must be escaped, and `\r` and `\n` are not
-// allowed. Sanitise input for safe use as reply text.
+// Sanitise input for safe use as reply text. RFC 7208, section 6.2 requires the
+// explanation string to be limited to ASCII. Also, according to libmilter API
+// documentation, `%` must be escaped, and `\r` and `\n` are not allowed.
 fn escape_reply_text(s: &str) -> Cow<'_, str> {
     fn is_regular_char(c: char) -> bool {
-        c.is_ascii_graphic() && !matches!(c, '%') || matches!(c, ' ' | '\t') || !c.is_ascii()
+        c.is_ascii_graphic() && !matches!(c, '%') || matches!(c, ' ' | '\t')
     }
 
     if s.chars().all(is_regular_char) {
@@ -444,46 +449,26 @@ fn add_headers(
     helo_host: Option<&str>,
     results: Vec<VerificationResult>,
 ) -> milter::Result<()> {
-    for result in results {
-        let VerificationResult { identity, spf_result, cause } = result;
-
-        // Iterate in reverse: if more than a single header is to be added, the
-        // resulting message reflects the order given in configuration.
-        for header_type in config.header().iter().rev() {
-            let header_field = match header_type {
-                HeaderType::ReceivedSpf => {
-                    ReceivedSpfHeader::new(
-                        &spf_result,
-                        cause.as_ref(),
-                        &hostname,
-                        ip,
-                        helo_host.as_deref(),
-                        &identity,
-                    )
-                    .to_header_field()
+    // Iterate in reverse: if more than a single header type is to be added, the
+    // resulting message reflects the order given in configuration.
+    for header_type in config.header().iter().rev() {
+        match header_type {
+            HeaderType::ReceivedSpf => {
+                for result in &results {
+                    let header = ReceivedSpfHeader::new(result, hostname, ip, helo_host);
+                    insert_header_field(context, config, id, header)?;
                 }
-                HeaderType::AuthenticationResults => {
-                    let authserv_id = authserv_id(config, &hostname);
-                    AuthenticationResultsHeader::new(
+            }
+            HeaderType::AuthenticationResults => {
+                if !results.is_empty() {
+                    let authserv_id = authserv_id(config, hostname);
+                    let header = AuthenticationResultsHeader::new(
                         authserv_id,
-                        &spf_result,
-                        cause.as_ref(),
-                        &identity,
+                        &results,
                         config.include_mailfrom_local_part(),
-                    )
-                    .to_header_field()
+                    );
+                    insert_header_field(context, config, id, header)?;
                 }
-            };
-
-            let header_name = header_field.name();
-
-            if config.dry_run() {
-                debug!("{}: adding {} header [dry run, not done]", id, header_name);
-                debug!("{}: {}", id, header_field);
-            } else {
-                debug!("{}: adding {} header", id, header_name);
-                debug!("{}: {}", id, header_field);
-                context.insert_header(1, header_name, &header_field.format_body())?;
             }
         }
     }
@@ -491,6 +476,26 @@ fn add_headers(
     Ok(())
 }
 
+fn insert_header_field(
+    context: &impl ActionContext,
+    config: &Config,
+    id: &str,
+    header_field: impl HeaderField,
+) -> milter::Result<()> {
+    let header_name = header_field.name();
+
+    if config.dry_run() {
+        debug!("{}: adding {} header [dry run, not done]", id, header_name);
+        debug!("{}: {}", id, header_field);
+    } else {
+        debug!("{}: adding {} header", id, header_name);
+        debug!("{}: {}", id, header_field);
+        context.insert_header(0, header_name, &header_field.format_body())?;
+    }
+
+    Ok(())
+}
+
 fn authserv_id<'a>(config: &'a Config, hostname: &'a str) -> &'a str {
     config.authserv_id().unwrap_or(hostname)
 }
@@ -511,8 +516,8 @@ mod tests {
     #[test]
     fn escape_reply_text_ok() {
         assert_eq!(escape_reply_text("a b%20c"), "a b%%20c");
-        assert_eq!(escape_reply_text("a\rb"), "a\\rb");
-        assert_eq!(escape_reply_text("a🟥b"), "a🟥b");
+        assert_eq!(escape_reply_text("\r"), "\\r");
+        assert_eq!(escape_reply_text("\x08🟥"), "\\u{8}\\u{1f7e5}");
     }
 
     #[test]
@@ -601,7 +606,7 @@ mod tests {
 
     impl From<(usize, &str, &str)> for InsertedHeader {
         fn from((index, name, value): (usize, &str, &str)) -> Self {
-            assert_eq!(index, 1);
+            assert_eq!(index, 0);
             Self::new(name, value)
         }
     }
@@ -840,7 +845,8 @@ mod tests {
             Some(ErrorReply::new(
                 "550",
                 "5.7.23",
-                "SPF validation failed: example.com explains: You, 敦文, are 100%% a fraud!",
+                "SPF validation failed: example.com explains: \
+                You, \\u{6566}\\u{6587}, are 100%% a fraud!",
             ))
         );
     }
@@ -889,7 +895,7 @@ mod tests {
         assert_eq!(status.unwrap(), Status::Accept);
         assert_eq!(
             *deleted_headers.borrow(),
-            vec![
+            [
                 DeletedHeader::new("Authentication-Results", 1),
                 DeletedHeader::new("Authentication-Results", 3),
             ]
@@ -1059,7 +1065,7 @@ mod tests {
     #[test]
     fn add_headers_for_all_results() {
         let config = Config::builder(SOCKET)
-            .header(HeaderType::AuthenticationResults)
+            .header(vec![HeaderType::AuthenticationResults, HeaderType::ReceivedSpf])
             .include_all_results(true)
             .build()
             .unwrap();
@@ -1100,14 +1106,25 @@ mod tests {
         assert_eq!(status.unwrap(), Status::Accept);
         assert_eq!(
             *inserted_headers.borrow(),
-            vec![
+            [
                 InsertedHeader::new(
-                    "Authentication-Results",
-                    "mail.example.org; spf=pass smtp.helo=mail.example.com",
+                    "Received-SPF",
+                    "pass (mail.example.org: domain mail.example.com has authorized\n\
+                    \thost 1.2.3.4) receiver=mail.example.org; client-ip=1.2.3.4;\n\
+                    \thelo=mail.example.com; identity=helo; mechanism=all"
+                ),
+                InsertedHeader::new(
+                    "Received-SPF",
+                    "pass (mail.example.org: domain of me@example.com has authorized\n\
+                    \thost 1.2.3.4) receiver=mail.example.org; client-ip=1.2.3.4;\n\
+                    \thelo=mail.example.com; envelope-from=\"me@example.com\"; identity=mailfrom;\n\
+                    \tmechanism=all"
                 ),
                 InsertedHeader::new(
                     "Authentication-Results",
-                    "mail.example.org; spf=pass smtp.mailfrom=example.com",
+                    "mail.example.org;\n\
+                    \tspf=pass smtp.helo=mail.example.com;\n\
+                    \tspf=pass smtp.mailfrom=example.com"
                 ),
             ]
         );
