@@ -8,7 +8,7 @@ use std::{
     str::FromStr,
 };
 use syslog::Facility;
-use viaspf::{record::ExplainString, Name, SpfResult};
+use viaspf::{record::ExplainString, Name, ParseNameError, SpfResult};
 
 #[derive(Clone, Copy, Debug, Default, Eq, Hash, PartialEq)]
 pub struct ParseStatusCodeError;
@@ -127,12 +127,12 @@ impl FromStr for EnhancedStatusCode {
 pub struct Header(Vec<HeaderType>);
 
 impl Header {
-    pub fn new(headers: Vec<HeaderType>) -> Result<Self, Vec<HeaderType>> {
+    pub fn new(header_types: Vec<HeaderType>) -> Result<Self, Vec<HeaderType>> {
         // Each header type must not appear more than once (ordered set).
-        if headers.len() == headers.iter().collect::<HashSet<_>>().len() {
-            Ok(Self(headers))
+        if header_types.len() == header_types.iter().collect::<HashSet<_>>().len() {
+            Ok(Self(header_types))
         } else {
-            Err(headers)
+            Err(header_types)
         }
     }
 
@@ -147,9 +147,15 @@ impl Default for Header {
     }
 }
 
+impl From<Vec<HeaderType>> for Header {
+    fn from(header_types: Vec<HeaderType>) -> Self {
+        Self(header_types)
+    }
+}
+
 impl From<HeaderType> for Header {
     fn from(header_type: HeaderType) -> Self {
-        Self(vec![header_type])
+        vec![header_type].into()
     }
 }
 
@@ -389,10 +395,47 @@ impl SkipEntry {
     }
 }
 
+impl FromStr for SkipEntry {
+    type Err = ParseNameError;
+
+    fn from_str(s: &str) -> Result<Self, Self::Err> {
+        let mut local_part = None;
+        let mut match_subdomains = false;
+
+        let domain = match s.rfind('@') {
+            Some(i) => {
+                // Use local-part as-is without further validation.
+                local_part = Some(s[..i].into());
+                &s[i + 1..]
+            }
+            None => match s.strip_prefix('.') {
+                Some(s) => {
+                    match_subdomains = true;
+                    s
+                }
+                None => s,
+            },
+        };
+
+        let domain = Name::domain(domain)?;
+
+        Ok(Self {
+            local_part,
+            domain,
+            match_subdomains,
+        })
+    }
+}
+
 #[derive(Clone, Debug, Default, Eq, PartialEq)]
 pub struct SkipSenders(HashSet<SkipEntry>);
 
 impl SkipSenders {
+    pub fn extended_with(mut self, other: Self) -> Self {
+        self.0.extend(other.0);
+        self
+    }
+
     pub fn includes(&self, sender: &str) -> bool {
         self.0.iter().any(|e| e.matches(sender))
     }
