@@ -2,7 +2,7 @@ use crate::config::{
     cli_opts::CliOptions,
     model::{
         DefinitiveHeloResults, EnhancedStatusCode, ExpModification, Header, HeaderType,
-        LogDestination, LogLevel, RejectResults, ReplyCode, SkipEntry, SkipSenders, SyslogFacility,
+        LogDestination, LogLevel, RejectResults, ReplyCode, SkipSenders, SyslogFacility,
         TrustedNetworks,
     },
     Config, ConfigBuilder, ConfigError,
@@ -17,12 +17,12 @@ use std::{
     path::{Path, PathBuf},
     time::Duration,
 };
-use viaspf::{record::ExplainString, Name, ParseNameError};
+use viaspf::record::ExplainString;
 
 // Implementation note: Remember not to use variable-sized types such as `usize`
 // in `RawConfig`. For this reason and others, the raw configuration does not
 // map onto `ConfigBuilder` exactly (eg `timeout_secs: u32` vs
-// `timeout: Duration`).
+// `timeout: Duration`, or `skip_senders`/`skip_senders_file`).
 
 /// Configuration parameter settings read from a file.
 #[derive(Clone, Debug, Default, Eq, PartialEq)]
@@ -48,6 +48,7 @@ struct RawConfig {
     permerror_status_code: Option<EnhancedStatusCode>,
     reject_helo_results: Option<RejectResults>,
     reject_results: Option<RejectResults>,
+    skip_senders: Option<SkipSenders>,
     skip_senders_file: Option<SkipSenders>,
     socket: Option<String>,
     softfail_reply_code: Option<ReplyCode>,
@@ -145,8 +146,14 @@ impl RawConfig {
         if let Some(v) = self.reject_results {
             builder = builder.reject_results(v);
         }
-        if let Some(v) = self.skip_senders_file {
-            builder = builder.skip_senders(v);
+        match (self.skip_senders, self.skip_senders_file) {
+            (Some(v), None) | (None, Some(v)) => {
+                builder = builder.skip_senders(v);
+            }
+            (Some(v1), Some(v2)) => {
+                builder = builder.skip_senders(v1.extended_with(v2));
+            }
+            (None, None) => {}
         }
         if let Some(v) = self.softfail_reply_code {
             builder = builder.softfail_reply_code(v);
@@ -271,7 +278,8 @@ pub enum ParseParamError {
     InvalidHeaderField(String),
     DuplicateHeaderField(HeaderType),
     ReadIncludedFile(PathBuf, String),
-    InvalidSkipEntry(String, PathBuf),
+    InvalidSkipEntry(String),
+    InvalidSkipEntryInFile(String, PathBuf),
 }
 
 impl Error for ParseParamError {}
@@ -298,7 +306,8 @@ impl Display for ParseParamError {
             Self::InvalidHeaderField(s) => write!(f, "invalid header field \"{}\"", s),
             Self::DuplicateHeaderField(h) => write!(f, "duplicate header field \"{}\"", h),
             Self::ReadIncludedFile(p, e) => write!(f, "failed to read file {}: {}", p.display(), e),
-            Self::InvalidSkipEntry(s, p) => write!(f, "invalid sender \"{}\" in {}", s, p.display()),
+            Self::InvalidSkipEntry(s) => write!(f, "invalid sender \"{}\"", s),
+            Self::InvalidSkipEntryInFile(s, p) => write!(f, "invalid sender \"{}\" in {}", s, p.display()),
         }
     }
 }
@@ -396,7 +405,8 @@ fn parse_config_entry(config: &mut RawConfig, key: &str, value: &str) -> ParsePa
         "permerror_status_code" => config.permerror_status_code = Some(value.parse_param()?),
         "reject_helo_results" => config.reject_helo_results = Some(value.parse_param()?),
         "reject_results" => config.reject_results = Some(value.parse_param()?),
-        "skip_senders_file" => config.skip_senders_file = Some(value.parse_param()?),
+        "skip_senders" => config.skip_senders = Some(value.parse_param()?),
+        "skip_senders_file" => config.skip_senders_file = Some(parse_skip_senders_file(value)?),
         "socket" => config.socket = Some(value.to_owned()),
         "softfail_reply_code" => config.softfail_reply_code = Some(value.parse_param()?),
         "softfail_reply_text" => config.softfail_reply_text = Some(value.to_owned()),
@@ -577,6 +587,47 @@ impl FromParamStr for TrustedNetworks {
     }
 }
 
+impl FromParamStr for SkipSenders {
+    fn from_param_str(s: &str) -> ParseParamResult<Self> {
+        let mut entries = HashSet::new();
+        for value in split_param_values(s) {
+            let value = value?;
+            let entry = value
+                .parse()
+                .map_err(|_| ParseParamError::InvalidSkipEntry(value.into()))?;
+            entries.insert(entry);
+        }
+        Ok(entries.into())
+    }
+}
+
+fn parse_skip_senders_file(s: &str) -> ParseParamResult<SkipSenders> {
+    let file_content = fs::read_to_string(s)
+        .map_err(|e| ParseParamError::ReadIncludedFile(s.into(), e.to_string()))?;
+
+    parse_skip_senders_file_content(&file_content, s)
+}
+
+fn parse_skip_senders_file_content(s: &str, file_name: &str) -> ParseParamResult<SkipSenders> {
+    let mut entries = HashSet::new();
+
+    for line in s.lines() {
+        let line = line.trim();
+
+        if is_ignored_line(line) {
+            continue;
+        }
+
+        let entry = line
+            .parse()
+            .map_err(|_| ParseParamError::InvalidSkipEntryInFile(line.into(), file_name.into()))?;
+
+        entries.insert(entry);
+    }
+
+    Ok(entries.into())
+}
+
 fn split_param_values(value: &str) -> impl Iterator<Item = ParseParamResult<&str>> {
     let value = value.trim();
 
@@ -598,62 +649,6 @@ fn split_param_values(value: &str) -> impl Iterator<Item = ParseParamResult<&str
     })
 }
 
-impl FromParamStr for SkipSenders {
-    fn from_param_str(s: &str) -> ParseParamResult<Self> {
-        let file_content = fs::read_to_string(s)
-            .map_err(|e| ParseParamError::ReadIncludedFile(s.into(), e.to_string()))?;
-
-        parse_skip_senders(&file_content, s)
-    }
-}
-
-fn parse_skip_senders(s: &str, file_name: &str) -> ParseParamResult<SkipSenders> {
-    let mut entries = HashSet::new();
-
-    for line in s.lines() {
-        let line = line.trim();
-
-        if is_ignored_line(line) {
-            continue;
-        }
-
-        let entry = parse_skip_entry(line)
-            .map_err(|_| ParseParamError::InvalidSkipEntry(line.into(), file_name.into()))?;
-
-        entries.insert(entry);
-    }
-
-    Ok(entries.into())
-}
-
-fn parse_skip_entry(s: &str) -> Result<SkipEntry, ParseNameError> {
-    let mut local_part = None;
-    let mut match_subdomains = false;
-
-    let domain = match s.rfind('@') {
-        Some(i) => {
-            // Use local-part as-is without further validation.
-            local_part = Some(s[..i].into());
-            &s[i + 1..]
-        }
-        None => match s.strip_prefix('.') {
-            Some(s) => {
-                match_subdomains = true;
-                s
-            }
-            None => s,
-        },
-    };
-
-    let domain = Name::domain(domain)?;
-
-    Ok(SkipEntry {
-        local_part,
-        domain,
-        match_subdomains,
-    })
-}
-
 fn is_ignored_line(line: &str) -> bool {
     let line = line.trim_start();
     line.is_empty() || line.starts_with('#')
@@ -662,7 +657,8 @@ fn is_ignored_line(line: &str) -> bool {
 #[cfg(test)]
 mod tests {
     use super::*;
-    use crate::config::model::{DefinitiveHeloResultKind, RejectResultKind};
+    use crate::config::model::{DefinitiveHeloResultKind, RejectResultKind, SkipEntry};
+    use viaspf::Name;
 
     #[test]
     fn socket_with_cli_options_override() {
@@ -755,20 +751,29 @@ mod tests {
     }
 
     #[test]
-    fn split_param_values_ok() {
-        assert!(split_param_values(" ").eq(vec![]));
-        assert!(split_param_values(" a").eq(vec![Ok("a")]));
-        assert!(split_param_values(" a,").eq(vec![Ok("a"), Err(ParseParamError::InvalidValue)]));
-        assert!(split_param_values(" ,a").eq(vec![Err(ParseParamError::InvalidValue), Ok("a")]));
-        assert!(split_param_values(" ,").eq(vec![
-            Err(ParseParamError::InvalidValue),
-            Err(ParseParamError::InvalidValue)
-        ]));
-        assert!(split_param_values(" a,b").eq(vec![Ok("a"), Ok("b")]));
+    fn parse_skip_senders_ok() {
+        let skip_senders = "Example.Com, me@Example.Com".parse_param::<SkipSenders>().unwrap();
+        assert_eq!(
+            skip_senders,
+            SkipSenders::from({
+                let mut results = HashSet::new();
+                results.insert(SkipEntry {
+                    local_part: None,
+                    domain: Name::domain("example.com").unwrap(),
+                    match_subdomains: false,
+                });
+                results.insert(SkipEntry {
+                    local_part: Some("me".into()),
+                    domain: Name::domain("example.com").unwrap(),
+                    match_subdomains: false,
+                });
+                results
+            })
+        );
     }
 
     #[test]
-    fn parse_skip_senders_ok() {
+    fn parse_skip_senders_file_content_ok() {
         let config = "
         # Domains
         Example.Com
@@ -778,7 +783,7 @@ mod tests {
         me@Example.Com
         ";
 
-        let skip_senders = parse_skip_senders(config, "unused").unwrap();
+        let skip_senders = parse_skip_senders_file_content(config, "unused").unwrap();
 
         assert_eq!(
             skip_senders,
@@ -803,4 +808,17 @@ mod tests {
             })
         );
     }
+
+    #[test]
+    fn split_param_values_ok() {
+        assert!(split_param_values(" ").eq(vec![]));
+        assert!(split_param_values(" a").eq(vec![Ok("a")]));
+        assert!(split_param_values(" a,").eq(vec![Ok("a"), Err(ParseParamError::InvalidValue)]));
+        assert!(split_param_values(" ,a").eq(vec![Err(ParseParamError::InvalidValue), Ok("a")]));
+        assert!(split_param_values(" ,").eq(vec![
+            Err(ParseParamError::InvalidValue),
+            Err(ParseParamError::InvalidValue)
+        ]));
+        assert!(split_param_values(" a,b").eq(vec![Ok("a"), Ok("b")]));
+    }
 }
